<!doctype html>

{% set app_registry_url = "http://aiidateam.github.io/aiida-registry" %}
{% set aiidalab_url = "https://www.materialscloud.org/aiidalab" %}

<html lang="en">

<head>
    <meta charset="utf-8">

    <title>{{title}}</title>
    <meta name="description" content="{{description}}">

    <link href="https://fonts.googleapis.com/css?family=Noto+Sans" rel="stylesheet">
    {% block css %}
    <link rel="stylesheet" href="static/css/style.css">
    <link rel="shortcut icon" type="image/png" href="static/favicon.png">
    {% endblock css %}

    <!--[if lt IE 9]>
    <script src="https://cdnjs.cloudflare.com/ajax/libs/html5shiv/3.7.3/html5shiv.js"></script>
  <![endif]-->
</head>

<body>
    {% block header %}
    <header id='entrytitle' style="background-color: #000; margin: 0; padding: 5px 20px 25px 20px;">
        <h1>
            <a href="{{ aiidalab_url }}">AiiDAlab</a> application registry
        </h1>
      </header>

    {% endblock header %}
  <div style="max-width: 1200px; margin: auto;">
    {% block body %} {% endblock body %}
  </div>
    <footer class="footer">
    {% block footer %}
    <hr>
    This page was auto-generated by the <a href="{{ aiidalab_url }}">AiiDAlab</a> application.
    {% endblock footer %}
    </footer>
</body>

</html>
